<?php
require_once("backend.inc.php");

/**
 * Ask the user to login via HTTP authentication.
 * Dies if the credentials are invalid.
 *
 * @return The numeric ID of the user.
 */
function ensure_login() {
	if (!isset($_SERVER['PHP_AUTH_USER'])) {
		// No credentials sent
		http_response_code(401); // Unauthorized
		header('WWW-Authenticate: Basic realm="'.addslashes(get_realm()).'"');
		die("Missing login");
	} else {
		// Credentials sent
		$user_id = login($_SERVER['PHP_AUTH_USER'], $_SERVER['PHP_AUTH_PW']);

		if ($user_id == -1) {
			// Invalid credentials
			header('WWW-Authenticate: Basic realm="'.addslashes(get_realm()).'"');
			http_response_code(401); // Unauthorized
			die("Invalid login");
		} else {
			// Valid credentials
			return $user_id;
		}
	}
}

/**
 * Checks whether a user has write access.
 * Dies if not.
 *
 * @param user_id The numeric ID of the user.
 */
function ensure_has_write_access($user_id) {
	if (!has_write_access($user_id)) {
		// No write access
		http_response_code(401); // Unauthorized
		header('WWW-Authenticate: Basic realm="'.addslashes(get_realm()).'"');
		die("User has no write access");
	}
}

/**
 * Checks the If-Match HTTP header field to ensure a unique version of a resource is present.
 * Expects the file's MD5 hash as the ETag.
 * Dies if the hash does not match.
 *
 * @param path The path to the file to check.
 */
function ensure_match($path) {
	// Ensure the data was not changed by another client by comparing hashes
	if (isset($_SERVER['HTTP_IF_MATCH']) and (trim(stripslashes($_SERVER['HTTP_IF_MATCH']), '"') != md5_file($path))) {
		http_response_code(412); // Precondition Failed
		die("Possible lost update (race condition)");
	}
}

/**
 * Prints the content of a file to the output stream. Returns silently if the file does not exist.
 * Sends the file's MD5 hash as the ETag.
 *
 * @param path The path to the file to print.
 */
function print_file($path) {
	if (!file_exists($path)) return;

	header("Cache-Control: no-cache");
	header("Pragma: no-cache");
	header("Content-Type: application/zip");
	header("Content-Disposition: attachment; filename=data.zip");
	header("Content-Length: ".filesize($path));
	header('ETag: "'.md5_file($path).'"');
	readfile($path);
}

/**
 * Stores the HTTP PUT input in a file in an atomic fashion.
 *
 * @param path The path to the file to store the data in.
 * @param quota The maximum size the file may reach before dying with a "Quota exceeded" message.
 */
function put_to_file($path, $quota) {
	// Write to temp file first to allow rollback in case of error
	$temp_path = tempnam(sys_get_temp_dir(), "0install-sync");

	$putdata = fopen("php://input", "r");
	$file = fopen($temp_path, "w");
    if (!$file) {
        http_response_code(500); // Internal Server Error
        die("Failed to create temp file");
    }

	$bytes_written = 0;
	while ($chunk = fread($putdata, 1024)) {
		$bytes_written += fwrite($file, $chunk);
		if ($bytes_written > $quota) {
			http_response_code(413); // Request Entity Too Large
			die("Quota exceeded");
		}
	}

	fclose($file);
	fclose($putdata);

	// Finish transaction as last step
    if (!rename($temp_path, $path)) {
        http_response_code(500); // Internal Server Error
        die("Failed to store file");
    }
}


$user_id = ensure_login();
$blob_type = $_GET['type'];
if ($blob_type != 'app-list' and $blob_type != 'app-list-machine') {
	http_response_code(400); // Bad Request
	die("Invalid blob type");
}

$blob_path = get_blob_path($user_id, $blob_type);
$blob_lock = fopen($blob_path.".lock", "a+");

switch ($_SERVER['REQUEST_METHOD']) {
	case "HEAD":
		// Retreive only headers without data; usually used to verify credentials
		header('Allow: HEAD, GET, PUT, DELETE');
		header("Cache-Control: no-cache");
		header("Pragma: no-cache");
		break;

	case "GET":
		// Get (shared) reader lock and then output file with tag
		flock($blob_lock, LOCK_SH);
		print_file($blob_path);
		flock($blob_lock, LOCK_UN);
		break;

	case "PUT":
		// Get (exclusive) write lock and then overwrite the file if the tag matches
		ensure_has_write_access($user_id);
		flock($blob_lock, LOCK_EX);
		ensure_match($blob_path);
		put_to_file($blob_path, get_blob_quota($user_id, $blob_type));
		flock($blob_lock, LOCK_UN);
		break;

	case "DELETE":
		// Get (exclusive) write lock and then delete the file if the tag matches
		ensure_has_write_access($user_id);
		flock($blob_lock, LOCK_EX);
		ensure_match($blob_path);
		unlink($blob_path);
		flock($blob_lock, LOCK_UN);
		break;

	default:
		http_response_code(405); // Method Not Allowed
		header('Allow: HEAD, GET, PUT, DELETE');
		die("Must use HTTP HEAD, GET, PUT or DELETE");
}
?>
